import os
import shutil
import subprocess
from pathlib import Path


__all__ = ["compile"]




def _detect_compute_capability(gpu_index: int | None = None) -> str:
"""Return compute capability like '80' (for sm_80). Picks first GPU unless index provided."""
cmd = [
"nvidia-smi",
"--query-gpu=compute_cap",
"--format=csv,noheader",
]
out = subprocess.check_output(cmd, text=True).strip().splitlines()
if not out:
raise RuntimeError("nvidia-smi returned no GPUs; is NVIDIA driver installed?")
line = out[gpu_index or 0].strip().replace(".", "")
if not line.isdigit():
raise RuntimeError(f"Unexpected compute_cap output: {line!r}")
return line




def compile(
sources_dir: str | os.PathLike = "../cuOPO3D",
output_exe: str | os.PathLike = "cuOPO3D",
dispersion: bool = True,
opt_level: int = 3,
gpu_index: int | None = None,
extra_nvcc: list[str] | None = None,
env: dict[str, str] | None = None,
) -> Path:
"""
Compile the CUDA solver with nvcc. Returns path to the executable.


Parameters
----------
sources_dir : path to directory containing cuOPO3D.cu
output_exe : executable name or path to create
dispersion : if True, add -DDISPERSION
opt_level : 0..3 optimization level
gpu_index : pick GPU index for compute capability detection
extra_nvcc : list of extra flags for nvcc
env : environment overrides for subprocess
"""
sources_dir = Path(sources_dir).resolve()
cu_file = sources_dir / "cuOPO3D.cu"
if not cu_file.exists():
raise FileNotFoundError(f"cu file not found: {cu_file}")


sm = _detect_compute_capability(gpu_index)


nvcc = shutil.which("nvcc")
if not nvcc:
raise RuntimeError("nvcc not found in PATH. Install CUDA toolkit or update PATH.")


cflags = [
nvcc,
str(cu_file),
"-O" + str(opt_level),
"-diag-suppress",
"177",
"-gencode=arch=compute_" + sm + ",code=sm_" + sm,
return exe_path


import os
import shutil
import json
import subprocess
import time
import glob
import re
import sys 

from .utils import load_config, save_config


def write_grid_on_file(json_path='config.json', cuh_path='headers/SetGrid.cuh'):
    with open(json_path, 'r') as f:
        config = json.load(f)
    # take values from file
    grid = config["grid"]["grid_points"]
    kernel = config["grid"]["kernel_dim"]

    NX = grid["NX"]
    NY = grid["NY"]
    NZ = grid["NZ"]
    NT = grid["NT"]
    BLKT = kernel["BLKT"]
    BLKX = kernel["BLKX"]
    BLKY = kernel["BLKY"]

    # set global variables
    content = (
        "#ifndef _SETGRIDCUH\n"
        "#define _SETGRIDCUH\n\n"
        f"const uint32_t NX = {NX};\n"
        f"const uint32_t NY = {NY};\n"
        f"const uint32_t NZ = {NZ};\n"
        f"#ifdef DISPERSION\n"
        f"const uint32_t NT = {NT};\n"
        f"#else\n"
        f"const uint32_t NT = 1;\n"
        f"#endif\n"
        f"const uint32_t BLKT = {BLKT};\n"
        f"const uint32_t BLKX = {BLKX};\n"
        f"const uint32_t BLKY = {BLKY};\n"
        f"const uint32_t SIZE = NX*NY*NT;\n"
        "\n\n#endif // _SETGRIDCUH\n"
    )

    # write the file
    with open(cuh_path, 'w') as f:
        f.write(content)
    print(f"File {cuh_path} filled with the corresponding grid values.")


def compile_cutwm(dispersion=False, twm_process="shg", compiler="nvcc", opt=3):
    """
    Compile the executable file cuTWM with flags.
    """
    import subprocess

    print("*-----------------------------------------------------------------*") 
    write_grid_on_file(json_path='config.json', cuh_path='headers/SetGrid.cuh')

    # Detect automatically compute capability
    compute_cap = subprocess.check_output(
        "nvidia-smi --query-gpu=compute_cap --format=csv,noheader | head -1 | tr -d '.'",
        shell=True
    ).decode().strip()

    flags = [
        compiler, "cuTWM.cu", "-w",
        f"-O{opt}",
        f"-gencode=arch=compute_{compute_cap},code=sm_{compute_cap}",
        f"-gencode=arch=compute_{compute_cap},code=compute_{compute_cap}", # f"-diag-suppress 177",
        "-lcufftw", "-lcufft", "-o", "cuTWM"
    ]

    if dispersion:
        flags.insert(2, "-DDISPERSION")

    if twm_process == "shg":
        flags.insert(3, "-DSHG")
    elif twm_process == "opg":
        flags.insert(3, "-DOPG")
    else:   
        raise RuntimeError("Compilation failed: invalid twm_process")

    print("*-----------------------------------------------------------------*")
    print("Compiling with:", " ".join(flags))
    result = subprocess.run(flags)
    if result.returncode != 0:
        raise RuntimeError("Compilation failed")
    print("Compilation succeeded.")



# ---------------- Logging: captura en tiempo real stdout/stderr de un comando ----------------
def run_and_tee(cmd, log_path, cwd=None, env=None):
    """
    Ejecuta `cmd` capturando TODO lo que imprime a terminal y guardándolo en `log_path`,
    mientras también lo muestra en pantalla en tiempo real.
    """
    log_dir = os.path.dirname(log_path)
    if log_dir:
        os.makedirs(log_dir, exist_ok=True)

    with open(log_path, "a", encoding="utf-8", buffering=1) as lf:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            bufsize=1,
            cwd=cwd,
            env=env
        )
        # Leer línea a línea y volcar a consola y archivo
        for line in proc.stdout:
            sys.stdout.write(line)
            lf.write(line)
        returncode = proc.wait()

    return returncode


def run_cutwm(config_path="config.json", exe_path="./cuTWM", log_path=None, cwd=None, env=None):
    """
    Run the main file with config.json.

    Si `log_path` es None, se genera automáticamente junto al config:
        <misma_carpeta>/<misma_base>.log
    """
    # Derivar log por defecto junto al JSON
    if log_path is None:
        cfg_abs = os.path.abspath(config_path)
        base, _ = os.path.splitext(cfg_abs)
        log_path = base + ".log"

    cmd = [exe_path, config_path]
    print("Running:", " ".join(cmd))
    rc = run_and_tee(cmd, log_path=log_path, cwd=cwd, env=env)
    if rc != 0:
        raise RuntimeError(f"Simulation failed (exit code {rc}). Log: {log_path}")
    print(f"\n\nSimulation finished. Log saved to: {log_path}\n\n")


def move_dat_files(dest_folder):
    """
    Move all .dat file from current dir to dest_folder.
    """
    # Create folder if does not exist
    if not os.path.exists(dest_folder):
        os.makedirs(dest_folder)

    # Sweep files in the current folder
    for file in os.listdir("."):
        if file.endswith(".dat"):
            src = os.path.join(".", file)
            dst = os.path.join(dest_folder, file)
            shutil.move(src, dst)
            print(f"Moved: {file} -> {dest_folder}")


# ---------- Helpers ----------
def _format_value(v, decimals=1) -> str:
    """
    Formatea el valor para nombres. Para floats, fija el número de decimales.
    Ej: 1   -> "1"
        1.0 -> "1.0" (si decimals=1)
        5.25 -> "5.3" (si decimals=1)
    """
    if isinstance(v, float):
        return f"{v:.{decimals}f}"
    return str(v)

def _sanitize_value_str(s: str) -> str:
    """
    Permite letras, números, punto, guion, guion_bajo, + y =.
    NO reemplaza el punto.
    """
    return re.sub(r"[^A-Za-z0-9.\-+=_]", "_", s)

def _value_str(val, decimals=1) -> str:
    """Formatea y sanitiza en un solo paso."""
    return f"{val:.{decimals}f}"


# ---------- Sweep ----------
def run_cutwm_sweep(config_path, param_path, values,
                    exe_path="./cuTWM",
                    out_dir_base="./results/sweep",
                    keep_config=True,
                    source_dir=".",              # de dónde recoger .dat si el exe no respeta output_directory
                    patterns=("*.dat",),         # qué patrones mover post-corrida
                    mtime_tolerance_s=1.0,       # tolerancia por resolución de timestamps
                    value_decimals=1             # decimales para el valor en nombres
                    ):
    """
    Barre un único parámetro y deja la salida de cada corrida en:
        <out_dir_base>/simulation_<leaf>_<valor>/

    - Si el ejecutable respeta cfg['simulation']['output_directory'], los .dat
      aparecerán directamente dentro de esa carpeta.
    - Si NO lo respeta y escribe en otra parte (p. ej. CWD), se intentará mover
      los .dat recién creados/modificados desde `source_dir`.

    Además, genera un .log con el MISMO nombre que el JSON temporal de la corrida,
    guardándolo en la misma carpeta (out_dir).
    """
    cfg = load_config(config_path)

    keys = param_path.split("/")
    leaf = keys[-1]

    os.makedirs(out_dir_base, exist_ok=True)

    for v in values:
        # 1) aplicar el valor en la config (nota: si tu ejecutable modifica el JSON en memoria,
        #    considera recargar cfg cada iteración con load_config(config_path))
        ptr = cfg
        for k in keys[:-1]:
            ptr = ptr[k]
        ptr[leaf] = v

        # 2) construir nombre y carpeta (con punto en floats)
        val_str = _value_str(v, decimals=value_decimals)
        folder_name = f"simulation_{leaf}_{val_str}"
        out_dir = os.path.join(out_dir_base, folder_name)
        os.makedirs(out_dir, exist_ok=True)

        # 3) actualizar metadatos de simulación
        base_name = cfg.get('simulation', {}).get('name', 'sim')
        sim_name = f"{base_name}_{leaf}_{val_str}"
        cfg['simulation']['name'] = sim_name
        cfg['simulation']['output_directory'] = out_dir

        # 4) guardar config dentro de la carpeta de salida
        temp_config = os.path.join(out_dir, f"config_{leaf}_{val_str}.json")
        save_config(cfg, temp_config)

        # 5) ejecutar, registrando también un log con el MISMO nombre base que el JSON
        #    (p.ej., config_pump_power_W_1.0.json -> config_pump_power_W_1.0.log)
        log_path = os.path.splitext(temp_config)[0] + ".log"
        t0 = time.time()
        run_cutwm(config_path=temp_config, exe_path=exe_path, log_path=log_path)

        # 6) si no hay .dat en out_dir, mover los recién generados desde source_dir
        has_dat_in_out = any(fn.lower().endswith(".dat") for fn in os.listdir(out_dir))
        if not has_dat_in_out:
            for pat in patterns:
                for f in glob.glob(os.path.join(source_dir, pat)):
                    try:
                        # mover archivos nuevos/modificados desde el inicio de la corrida
                        if os.path.getmtime(f) >= (t0 - mtime_tolerance_s):
                            dst = os.path.join(out_dir, os.path.basename(f))
                            if os.path.exists(dst):
                                base, ext = os.path.splitext(dst)
                                dst = f"{base}__{leaf}_{val_str}{ext}"
                            shutil.move(f, dst)
                    except FileNotFoundError:
                        pass

        # 7) limpieza opcional del config
        if not keep_config:
            try:
                os.remove(temp_config)
            except OSError:
                pass



import json

def load_config(config_path):
    with open(config_path, "r") as f:
        return json.load(f)

def save_config(cfg, out_path):
    with open(out_path, "w") as f:
        json.dump(cfg, f, indent=2)



import numpy as np
import matplotlib.pyplot as plt
import os
import json
import imageio
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

def plot_signal_intensity(real_file, imag_file):
    real = np.loadtxt(real_file)
    imag = np.loadtxt(imag_file)
    intensity = real**2 + imag**2
    plt.plot(intensity)
    plt.xlabel("Grid point / Time")
    plt.ylabel("Intensity |E|²")
    plt.title("Signal Field Intensity")
    plt.show()


def plot_XY_profile(real_path, imag_path,
                    Lx_um=1000.0, Ly_um=1000.0,
                    save_fig=True, out_png="fig_name.png"):
    """
    Read two .dat (real e imaginary parts) as 2D-matrices (NY, NX),
    compute |A|^2 and plot x-y in micron scale.
    """
    real = np.loadtxt(real_path)
    imag = np.loadtxt(imag_path)

    if real.shape != imag.shape:
        raise ValueError(f"Shape mismatch: real {real.shape} vs imag {imag.shape}")

    NY, NX = real.shape
    
    A2 = real**2 + imag**2

    dx_um = Lx_um / (NX - 1) if NX > 1 else float("nan")
    dy_um = Ly_um / (NY - 1) if NY > 1 else float("nan")

    plt.figure(figsize=(7, 6))
    extent = [0, Lx_um, 0, Ly_um]
    im = plt.imshow(A2, origin="lower", extent=extent, aspect="equal")
    plt.xlabel("x (μm)")
    plt.ylabel("y (μm)")
    plt.title(f"|ψ|²   NX={NX}, NY={NY},   dx={dx_um:.3g} μm, dy={dy_um:.3g} μm")
    plt.colorbar(im, label="Intensity (arb. units)")
    plt.tight_layout()
    if save_fig:
        plt.savefig(out_png, dpi=150, bbox_inches="tight")
    plt.show()
    return {"NX": NX, "NY": NY, "dx_um": dx_um, "dy_um": dy_um, "out_png": out_png}


def plot_fields_from_slices(
    files_folder,
    config_json_path,
    n_dict,        # {'pump': n_p, 'signal': n_s, 'idler': n_i}
    cmap='inferno',
    image_dir="intensity_cut_YZ/",
    image_name="intensity_profile",
    image_extension=".png"
):
    """
    Visualiza la propagación de los campos habilitados en config.json

    Args:
        files_folder (str): Carpeta con archivos .dat
        config_json_path (str): Ruta al archivo config.json
        n_dict (dict): {'pump': n_p, 'signal': n_s, 'idler': n_i}
        aspect_ratio (float): Relación de aspecto para cada panel
        cmap (str): Colormap
    """

    # --- Leer config.json ---
    with open(config_json_path, "r") as f:
        cfg = json.load(f)
    save_mode = cfg.get("save_mode", {})
    grid_points = cfg.get("grid", {}).get("grid_points", {})
    NX = int(grid_points.get("NX", 32))
    NY = int(grid_points.get("NY", 32))
    NZ = int(grid_points.get("NZ", 1))

    # --- Qué campos graficar ---
    field_list = [field for field in ('pump', 'signal', 'idler') if save_mode.get(f"save_{field}", False)]
    if not field_list:
        print("No hay campos seleccionados para graficar según config.json.")
        return

    # --- Leer archivos de cada campo ---
    E_fields = {}
    for field in field_list:
        E_cmplx = np.zeros((NZ, NY, NX), dtype=np.complex128)
        for z in range(NZ):
            for part in ('r', 'i'):
                fname = os.path.join(files_folder, f"{field}_output_XY_z_slice_{z}_{part}.dat")
                if not os.path.isfile(fname):
                    raise FileNotFoundError(f"No se encontró {fname}")
                data = np.loadtxt(fname)
                if part == 'r':
                    E_cmplx[z] = data
                else:
                    E_cmplx[z] += 1j * data
        E_fields[field] = E_cmplx

    # --- Graficar ---
    x_idx = NX // 2
    y = np.arange(NY)
    z_pos = np.arange(NZ)

    fig, axes = plt.subplots(len(field_list), 1, figsize=(5, 5), constrained_layout=True)

    if len(field_list) == 1:
        axes = [axes]

    for i, field in enumerate(field_list):
        n = n_dict[field]
        E = E_fields[field]
                # Physical constants in user units:
        C0 = 299792458 * 1e6 / 1e12        # Speed of light in vacuum [μm/ps]
        EPS0 = 8.8541878128e-12 * 1e12 / 1e6 # Vacuum permittivity [W·ps/V²·μm]
        intensity_Wum2 = C0*EPS0*n*np.abs(E[:, :, x_idx])**2 / 2.0
        intensity_Wcm2 = intensity_Wum2 * 1e8
        ax = axes[i]
        im = ax.imshow(
            intensity_Wcm2.T,
            origin='lower',
            aspect='auto',         # <-- esto centra SIEMPRE
            extent=[0, NZ-1, 0, NY-1],
            cmap=cmap
        )
        ax.set_title(f"{field.capitalize()} field")
        ax.set_xlabel('z-slice')
        ax.set_ylabel('y')
        cb = fig.colorbar(im, ax=ax, fraction=0.045)
        cb.set_label("Intensity (W/cm²)")
    plt.suptitle("Propagación de los campos (corte en x=NX/2)")

    # Ensure output directory exists
    output_image_dir = os.path.dirname(image_dir)
    if output_image_dir and not os.path.exists(image_dir):
        os.makedirs(image_dir)
    fig.savefig(image_dir+image_name+image_extension, dpi=150, bbox_inches='tight')
    plt.show()
    return


def make_field_propagation_gif(
    files_folder,
    config_json_path,
    n_dict,           # {'pump': n_p, 'signal': n_s, 'idler': n_i}
    gif_path,
    cmap='inferno',
    fps=10,
    vmin=None,
    vmax=None
):
    """
    Generates a GIF showing the progression of field intensity (x-y plane) at each slice along the crystal
    for each field enabled in config.json. Each frame is a horizontal row of 2D intensity maps (one per field).

    Args:
        files_folder (str): Path to the folder containing the .dat files.
        config_json_path (str): Path to config.json.
        n_dict (dict): Refractive indices for each field. Example: {'pump': 2.23, 'signal': 2.21, 'idler': 2.19}
        gif_path (str): Output path for the GIF file (should end with .gif).
        cmap (str): Colormap to use for intensity images.
        fps (int): Frames per second for the GIF animation.
        vmin, vmax (float or None): Set fixed min/max for color scale. If None, scale automatically for each field.
    """

    # --- Load config.json ---
    with open(config_json_path, "r") as f:
        cfg = json.load(f)

    save_mode = cfg.get("save_mode", {})
    grid_points = cfg.get("grid", {}).get("grid_points", {})
    NX = int(grid_points.get("NX", 32))
    NY = int(grid_points.get("NY", 32))
    NZ = int(grid_points.get("NZ", 1))
    crystal = cfg.get("crystal", {})
    Lcr = float(crystal.get("dimensions", {}).get("Lcr", 10000.0))  # in micrometers

    # --- Which fields to plot? ---
    field_list = [field for field in ('pump', 'signal', 'idler') if save_mode.get(f"save_{field}", False)]
    if not field_list:
        print("No fields selected for plotting in config.json.")
        return

    dz = Lcr / (NZ - 1) if NZ > 1 else 0

    # --- Precompute color limits for all fields over all slices (for consistent colorbar) ---
    if vmin is None or vmax is None:
        all_intensities = {field: [] for field in field_list}
        for z in range(NZ):
            for field in field_list:
                fname_real = os.path.join(files_folder, f"{field}_output_XY_z_slice_{z}_r.dat")
                fname_imag = os.path.join(files_folder, f"{field}_output_XY_z_slice_{z}_i.dat")
                real = np.loadtxt(fname_real)
                imag = np.loadtxt(fname_imag)
                E = real + 1j * imag
                # Physical constants in user units:
                C0 = 299792458 * 1e6 / 1e12        # Speed of light in vacuum [μm/ps]
                EPS0 = 8.8541878128e-12 * 1e12 / 1e6 # Vacuum permittivity [W·ps/V²·μm]
                intensity_Wum2 = C0*EPS0*n_dict[field]*np.abs(E)**2 / 2.0  # (NY, NX)
                intensity_Wcm2 = intensity_Wum2 * 1e8
                all_intensities[field].append(intensity_Wcm2)
        # Flatten and find global min/max for all fields
        flat_all = np.concatenate([np.ravel(arr) for field in field_list for arr in all_intensities[field]])
        vmin = float(np.nanmin(flat_all))
        vmax = float(np.nanmax(flat_all))

    # --- Ensure output directory exists ---
    output_dir = os.path.dirname(gif_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # --- Create and save frames ---
    images = []
    for z in range(NZ):
        # Create figure and attach Agg canvas
        fig = Figure(figsize=(6*len(field_list), 5), constrained_layout=True)
        FigureCanvas(fig)
        axes = fig.subplots(1, len(field_list))
        if len(field_list) == 1:
            axes = [axes]

        for i, field in enumerate(field_list):
            fname_real = os.path.join(files_folder, f"{field}_output_XY_z_slice_{z}_r.dat")
            fname_imag = os.path.join(files_folder, f"{field}_output_XY_z_slice_{z}_i.dat")
            real = np.loadtxt(fname_real)
            imag = np.loadtxt(fname_imag)
            E = real + 1j * imag
            # Physical constants in user units:
            C0 = 299792458 * 1e6 / 1e12        # Speed of light in vacuum [μm/ps]
            EPS0 = 8.8541878128e-12 * 1e12 / 1e6 # Vacuum permittivity [W·ps/V²·μm]
            intensity_Wum2 = C0*EPS0*n_dict[field]*np.abs(E)**2 / 2.0  # shape (NY, NX)
            intensity_Wcm2 = intensity_Wum2 * 1e8

            ax = axes[i]
            im = ax.imshow(
                intensity_Wcm2,
                origin='lower',
                aspect='auto',
                extent=[0, NX-1, 0, NY-1],
                cmap=cmap,
                vmin=vmin,
                vmax=vmax
            )
            ax.set_title(f"{field.capitalize()} field")
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            cb = fig.colorbar(im, ax=ax, fraction=0.045)
            cb.set_label("Intensity (W/cm²)")

        z_um = z * dz
        fig.suptitle(f"z = {z_um:.1f} μm")

        # Draw and convert figure to numpy RGB array
        fig.canvas.draw()
        image = np.asarray(fig.canvas.buffer_rgba())[:, :, :3]
        images.append(image)
        plt.close('all')
    
    # --- Save GIF ---
    # Ensure output directory exists
    output_dir = os.path.dirname(gif_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    imageio.mimsave(gif_path, images, fps=fps)
    print(f"GIF saved to {gif_path}")


# -*- coding: utf-8 -*-
from __future__ import annotations
import json
import os
import glob
from typing import Any, Dict, Optional, Tuple
import numpy as np

# Physical constants in user units:
C0 = 299792458 * 1e6 / 1e12        # Speed of light in vacuum [μm/ps]
EPS0 = 8.8541878128e-12 * 1e12 / 1e6 # Vacuum permittivity [W·ps/V²·μm]

# Output field file names
FIELD_FILES = {
    "p": ("pump_output_XY_r.dat",   "pump_output_XY_i.dat"),
    "s": ("signal_output_XY_r.dat", "signal_output_XY_i.dat"),
    "i": ("idler_output_XY_r.dat",  "idler_output_XY_i.dat"),
}

# -------------------------
# JSON utilities
# -------------------------
def read_from_json(json_path: str, key: str, default: Any = None) -> Any:
    """
    Reads a value from a JSON file using a dotted key (supports nested keys like 'a.b.c').
    If the key does not exist and default is given, returns default. Otherwise raises KeyError.
    """
    with open(json_path, "r") as f:
        data = json.load(f)

    if "." not in key:
        if key in data:
            return data[key]
        if default is not None:
            return default
        raise KeyError(f"Key '{key}' not found in {json_path}")

    cur = data
    parts = key.split(".")
    for p in parts:
        if isinstance(cur, dict) and p in cur:
            cur = cur[p]
        else:
            if default is not None:
                return default
            raise KeyError(f"Key '{key}' not found in {json_path}")
    return cur

def _auto_find_config_json(folder: str) -> str:
    """
    Automatically finds the unique config_*.json file in the folder.
    """
    candidates = sorted(glob.glob(os.path.join(folder, "config_*.json")))
    if len(candidates) == 0:
        raise FileNotFoundError(f"No config_*.json found in {folder}")
    if len(candidates) > 1:
        raise RuntimeError(f"Multiple config_*.json files found in {folder}: {candidates}")
    return candidates[0]


def _grid_from_json(json_path: str) -> Tuple[int, int, float, float]:
    """
    Reads NX, NY, LX(μm), LY(μm) from the JSON file.
    Returns: NX, NY, dx (μm), dy (μm).
    """
    NX = int(read_from_json(json_path, "grid.grid_points.NX"))
    NY = int(read_from_json(json_path, "grid.grid_points.NY"))
    LX_um = float(read_from_json(json_path, "crystal.dimensions.LX"))
    LY_um = float(read_from_json(json_path, "crystal.dimensions.LY"))
    dx = LX_um / (NX-1)     # grid spacing in μm
    dy = LY_um / (NY-1)     # grid spacing in μm
    return NX, NY, dx, dy

# -------------------------
# Field loading
# -------------------------
def _load_complex_matrix(real_path: str, imag_path: str, shape: Tuple[int, int]) -> np.ndarray:
    """
    Loads real/imaginary matrices (plain text), reshapes to (NY, NX), and returns a complex array.
    Tries C order first; if it fails, tries Fortran order.
    """
    Er = np.loadtxt(real_path, dtype=np.float64)
    Ei = np.loadtxt(imag_path, dtype=np.float64)
    if Er.size != Ei.size:
        raise ValueError(f"Size mismatch between {os.path.basename(real_path)} and {os.path.basename(imag_path)}")
    NY, NX = shape
    try:
        Er = Er.reshape(NY, NX, order="C")
        Ei = Ei.reshape(NY, NX, order="C")
    except ValueError:
        Er = Er.reshape(NY, NX, order="F")
        Ei = Ei.reshape(NY, NX, order="F")
    return Er + 1j * Ei


def _maybe_load_field(folder: str, key: str, NX: int, NY: int) -> Optional[np.ndarray]:
    """
    Loads a complex field (pump, signal, or idler) from files if both real and imaginary parts exist.
    Returns None if either file is missing.
    """
    r_name, i_name = FIELD_FILES[key]
    r_path = os.path.join(folder, r_name)
    i_path = os.path.join(folder, i_name)
    if not (os.path.isfile(r_path) and os.path.isfile(i_path)):
        return None
    return _load_complex_matrix(r_path, i_path, shape=(NY, NX))

# -------------------------
# Optical power calculation
# -------------------------
def _intensity(E: np.ndarray, n: float, field_scale: float = 1.0) -> np.ndarray:
    """
    Computes the intensity: I = 0.5 * n * eps0 * c * |E|^2 [W/μm²]
    E should be in V/μm if field_scale=1.0, or scale accordingly.
    """
    E_vpm = E * field_scale
    return 0.5 * n * EPS0 * C0 * np.abs(E_vpm)**2


def _power(I: np.ndarray, dx: float, dy: float) -> float:
    """
    Integrates intensity over the (x, y) plane to obtain the total power in [W].
    dx and dy must be given in μm.
    """
    return float(I.sum() * dx * dy)


def compute_powers_from_folder(
    folder: str,
    n_p: float,
    n_s: float,
    n_i: float,
    json_path: Optional[str] = None,
    field_scale: float = 1.0
) -> Dict[str, float]:
    """
    Computes the output power for all present fields (pump, signal, idler) in the specified folder.

    Args:
      folder: Path to the simulation results folder.
      n_p, n_s, n_i: Refractive indices for pump, signal, idler.
      json_path: Path to the simulation config JSON. If None, it is auto-detected in the folder.
      field_scale: Multiply the electric field by this factor to convert it to V/μm. 
                   Use 1.0 if already in V/μm. Use 1e6 if in V/m.

    Returns:
      Dictionary with powers in W for present fields: e.g., {"p": ..., "s": ..., "i": ...}
    """
    if json_path is None:
        json_path = _auto_find_config_json(folder)

    NX, NY, dx, dy = _grid_from_json(json_path)

    out: Dict[str, float] = {}
    # Pump
    Ep = _maybe_load_field(folder, "p", NX, NY)
    if Ep is not None:
        Ip = _intensity(Ep, n=n_p, field_scale=field_scale)
        out["p"] = _power(Ip, dx, dy)

    # Signal
    Es = _maybe_load_field(folder, "s", NX, NY)
    if Es is not None:
        Is = _intensity(Es, n=n_s, field_scale=field_scale)
        out["s"] = _power(Is, dx, dy)

    # Idler
    Ei = _maybe_load_field(folder, "i", NX, NY)
    if Ei is not None:
        Ii = _intensity(Ei, n=n_i, field_scale=field_scale)
        out["i"] = _power(Ii, dx, dy)

    return out




from setuptools import setup, find_packages

setup(
    name="pycutwm",
    version="0.1.0",
    description="Modelling tools for package cuTWM(3+1)D",
    author="Alfredo Daniel Sanchez",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "matplotlib",
        "imageio"
    ],
    python_requires=">=3.8",
)



import numpy as np

def n_mgoppln(wavelength, temperature):
    """
    Sellmeier equation for MgO:PPLN.
    Parameters:
        wavelength: Wavelength in microns
        temperature: Temperature in Celsius
    Returns:
        Refractive index
    """
    f = (temperature - 24.5) * (temperature + 570.82)
    a1 = 5.756
    a2 = 0.0983
    a3 = 0.2020
    a4 = 189.32
    a5 = 12.52
    a6 = 1.32e-2
    b1 = 2.860e-6
    b2 = 4.700e-8
    b3 = 6.113e-8
    b4 = 1.516e-4

    G1 = a1 + b1 * f
    G2 = a2 + b2 * f
    G3 = a3 + b3 * f
    G4 = a4 + b4 * f

    L2 = wavelength ** 2
    n2 = G1 + G2 / (L2 - G3 ** 2) + G4 / (L2 - a5 ** 2) - a6 * L2
    return np.sqrt(n2)

def n_ppln(wavelength, temperature):
    """
    Sellmeier equation for PPLN (from Gayer et al.).
    Parameters:
        wavelength: Wavelength in microns
        temperature: Temperature in Celsius
    Returns:
        Refractive index
    """
    f = (temperature - 24.5) * (temperature + 570.82)
    a1 = 5.35583
    a2 = 0.100473
    a3 = 0.20692
    a4 = 100.0
    a5 = 11.34927
    a6 = 1.5334e-2
    b1 = 4.629e-6
    b2 = 1.1685e-7
    b3 = 3.9046e-8
    b4 = 1.6762e-4

    G1 = a1 + b1 * f
    G2 = a2 + b2 * f
    G3 = a3 + b3 * f
    G4 = a4 + b4 * f

    L2 = wavelength ** 2
    n2 = G1 + G2 / (L2 - G3 ** 2) + G4 / (L2 - a5 ** 2) - a6 * L2
    return np.sqrt(n2)

def n_mgospplt(wavelength, temperature):
    """
    Sellmeier equation for MgO:SPP-LT (from Gayer et al.).
    Parameters:
        wavelength: Wavelength in microns
        temperature: Temperature in Celsius
    Returns:
        Refractive index
    """
    f = (temperature - 24.5) * (temperature + 570.82)
    a1 = 5.113
    a2 = 0.0996
    a3 = 0.2102
    a4 = 189.69
    a5 = 12.48
    a6 = 1.32e-2
    b1 = 2.767e-6
    b2 = 3.728e-8
    b3 = 5.290e-8
    b4 = 1.275e-4

    G1 = a1 + b1 * f
    G2 = a2 + b2 * f
    G3 = a3 + b3 * f
    G4 = a4 + b4 * f

    L2 = wavelength ** 2
    n2 = G1 + G2 / (L2 - G3 ** 2) + G4 / (L2 - a5 ** 2) - a6 * L2
    return np.sqrt(n2)

def n_zgp(wavelength, pol):
    """
    Sellmeier equation for ZnGeP2 (ZGP).
    Parameters:
        wavelength: Wavelength in microns
        pol: 'o' (ordinary) or 'e' (extraordinary) polarization
    Returns:
        Refractive index
    """
    L2 = wavelength ** 2
    if pol == 'o':
        A = 8.0409
        B = 1.68625
        CC = 0.40824
        D = 1.2880
        E = 611.05
    elif pol == 'e':
        A = 8.0929
        B = 1.8649
        CC = 0.41468
        D = 0.84052
        E = 452.05
    else:
        raise ValueError("Polarization must be 'o' or 'e'")

    n2 = A + (B * L2) / (L2 - CC) + (D * L2) / (L2 - E)
    return np.sqrt(n2)
